# frozen_string_literal: true

# Legende
# Les script a plusieurs parties indique que les bash (Batch dans l'app)
# peuvent s'initier avec differents types de source
# Le besoin est donc d'isoler la logique associee au modele actuel
# ne pas oublier d'indexer les champs simples

# TODO: Réflechir au set de housing_requests_ids (validate main_applicant)
#
#
module Hermes
  # Applicant class
  class Applicant
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::Status
    include ::Concerns::Geolocalizable
    include ::Concerns::NotificationCenter::Targetable
    include ::Concerns::NotificationCenter::Notifiable
    
    hermes_tools 'revenues'
    
    field :test_created_at, type: Date
    
    field :temporary_id, type: String
    
    field :nud, type: String
    field :sne_profile_creation_date, type: DateTime
    field :ss_id, type: String
    
    field :first_name, type: String
    field :last_name, type: String
    field :patronymic_name, type: String
    field :nationality, type: String
    field :sms_notification, type: Boolean
    field :email, type: String
    field :email_advances_assistant, type: String
    field :event_notification, type: Boolean
    
    field :cellphone, type: String
    field :homephone, type: String
    field :workphone, type: String
    
    field :civility, type: String
    
    field :birth_date, type: Date
    field :birth_plus_30, type: Date
    
    field :work_contract_type, type: String
    field :profession_details, type: Hash
    field :has_other_employers, type: Boolean
    field :main_employer_name, type: String
    field :main_employer_siret, type: String
    field :workplace_postal_code, type: String
    field :workplace_locality, type: String
    field :foreign_workplace_postal_code, type: String
    field :foreign_workplace_locality, type: String
    field :foreign_workplace_country, type: String
    
    field :year_minus_one, type: String
    field :amount_year_minus_one, type: Float
    field :year_minus_two, type: String
    field :amount_year_minus_two, type: Float
    field :ressources_details, type: Hash
    field :monthly_incomes, type: Hash
    field :monthly_charges, type: Hash
    field :tax_revenues, type: Float
    field :apl_amount, type: Float
    field :family_situation, type: String
    field :family_situation_validated, type: Boolean
    
    field :asked_children_nb, type: Integer
    field :valid_children_nb, type: Integer
    
    field :situation_tax_treshold, type: Integer
    field :asked_locataire_hlm, type: Boolean
    field :valid_locataire_hlm, type: Boolean
    field :valid_priority_public, type: Boolean
    field :asked_priority_public, type: Boolean
    
    field :single_parent, type: Boolean
    field :priority_public, type: Boolean
    field :mobility, type: Boolean
    field :mutation, type: Boolean
    
    field :completed_requests_nb, type: Integer
    field :on_going_request_nb, type: Integer
    
    field :actual_housing_details, type: Hash, default: {}
    field :postal_details, type: Hash, default: {}
    field :is_owner, type: Boolean
    field :owned_housing_details, type: Hash, default: {}
    
    field :disability, type: String
    field :renewal_date, type: DateTime
    field :has_valid_year_minus_two, type: Boolean
    field :eligibility_query, type: String, default: '{"is_public":true}' # JSON formated as string
    
    field :main_user_id, type: BSON::ObjectId
    field :workplace_insee_code, type: String
    field :last_xml_flux_36
    #10095 ajout de nouveaux champs cerfa v4
    field :unemployment_period, type: Boolean
    field :unemployment_period_date_end, type: Date
    field :command_to_leave_date, type: Date
    field :pro_or_home_tel_number, type: String

    alias_attribute :district, :city
    
    common_indexes *%i[
      created_at
      updated_at
      nud
      sne_profile_creation_date
      ss_id
      first_name
      last_name
      patronymic_name
      nationality
      email
      event_notification
      cellphone
      civility
      birth_date
      work_contract_type
      has_other_employers
      year_minus_one
      amount_year_minus_one
      year_minus_two
      amount_year_minus_two
      ressources_details
      tax_revenues
      apl_amount
      family_situation
      situation_tax_treshold
      priority_public
      mobility
      mutation
      completed_requests_nb
      on_going_request_nb
      is_owner
      main_user_id
    ]
    
    attr_accessor :should_update_eligibility
    
    # temporary fields to handle user side effects
    field :name_will_change, type: Boolean, default: false
    field :email_will_change, type: Boolean, default: false
    
   # validates :main_employer_siret, length: {is: 14}, allow_blank: true
    has_many(
      :housing_requests,
      class_name: 'Hermes::HousingRequest'
    )
    has_many(
      :notifications,
      class_name: '::NotificationCenter::Notification',
      foreign_key: :targetable_id,
      dependent: :destroy,
      order: 'created_at DESC'
    )

    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    writer_auth_roles 'collaborateur entreprise', "collaborateur offre logement", 'superadministrateur'
    
    # validates_presence_of :first_name, :last_name
    
    ELLIGIBILY_FIELDS = %w[
      mutation
      mobility
      priority_public
      single_parent
    ].concat((1..12).map {|i| "treshold_category_#{i}"})
    
    before_create :before_create_actions
    before_save :before_actions # https://dev.hermes.openstack.local/admin/families/5d1239624ac06e4ca38c1670/mapping
    after_save :after_actions
    
    Applicant.add_status(:single)
    
    before_destroy :before_destroy_actions

    ###############################
    #### Notification #############
    def self.targets_selector(applicant, opt = {})
      [applicant]
    end

    def notification_mailer
      ::Hermes::NotificationCenter::HermesMailer
    end

    def notifiable_title(key, opt={})
      I18n.t("#{key}.title")
    end

    def notifiable_body(key, opt={})
      I18n.t("#{key}.body")
    end

    def notifiable_path
      "https://als-hermes-int.be-idys.com/housing_applications/#{id}"
    end
    ###############################
    ###############################
    
    def department
      postal_code[0..1] if postal_code
    end
    
    def before_destroy_actions
      cleanup_user
      destroy_applicant_link
      clear_coapplicants
      destroy_required_documents
    end
    
    def destroy_applicant_link
      applicant_links = Hermes::ApplicantLink.where('$or' => [{applicant_ids: id.to_s}, {applicant_ids: id}])
      applicant_links.each do |al|
        if al.applicant_ids.length > 1
          
          al.applicant_ids.delete(id.to_s)
          al.applicant_ids.delete(id)
          al.save
        else
          al.destroy
        end
      end
    end

    def clear_coapplicants
      Hermes::HousingRequest.where(applicant_ids: id.to_s).each do |hr|
        hr.remove_coapplicant(id)
        hr.save
      end 
    end 
    
    def destroy_required_documents
      Hermes::RequiredDocument.where(applicant_id: id).destroy_all
    end
    
    def cleanup_user
      if user && (user.main_applicant_id == id)
        errors.add(:base, 'Cannot destroy my applicant')
        false
        # self.user.set(main_applicant_id: nil)
      end
    end
    
    def links(type: nil)
      links = Hermes::ApplicantLink.where(applicant_ids: id.to_s)
      links = links.where(type: type) if type.present?
      Hermes::Applicant.where(:id.in => (links.pluck(:applicant_ids).flatten.compact.uniq - [id.to_s]))
    end
    
    def link_types_for(applicant_id)
      Hermes::ApplicantLink.where(:$and => [{applicant_ids: id.to_s}, {applicant_ids: applicant_id.to_s}]).pluck(:type).uniq
    end
    
    def first_active_request
      @first_active_request ||= Hermes::HousingRequest.find_by(main_applicant_id: self.id, is_active: true)
    end
    
    def coapplicant_first_active_request
      @coapplicant_first_active_request ||= Hermes::HousingRequest.order_by(created_at: :desc).find_by(is_active: true, applicant_ids: self.id)
    end 

    def coapplicants
      first_active_request&.coapplicants
    end
    
    def coapplicant_ids
      first_active_request&.coapplicants.map {|a| a[:id]}
    end
    
    def monthly_revenues
      (monthly_incomes || {}).values_fsum - (monthly_charges || {}).values_fsum
    end
    
    def update_birth_dependencies
      return if birth_date.blank?
      
      self.should_update_eligibility = true
      self.birth_plus_30 = birth_date + 30.years
    end
    
    def update_hr_monthly_revenues_coapplicant
      if  hrs = Hermes::HousingRequest.where(is_active: true, applicant_ids: self.id.to_s)
        hrs.each do |hr|
          hr.update_monthly_revenues
          hr.save if hr.monthly_revenues_changed?
        end 
      end 
    end 
    
    def check_monthly_revenues_dependency
      cond = monthly_incomes_changed? || monthly_charges_changed?
      if first_active_request && cond
        first_active_request.monthly_revenues_will_change = true
      end
    end

    def check_category_dependency
      if first_active_request && family_situation_changed?
        first_active_request.category_may_change = true
      end

      hrs = Hermes::HousingRequest.where(applicant_ids: id.to_s, is_active: true)
      hrs.each do |hr|
        hr.update_category
        hr.save if hr.category_type_changed?
      end
    end
    
    def check_eligibility_dependency
      # TODO: set to false after process
      cond = (changes.keys & ELLIGIBILY_FIELDS).size.positive?
      if first_active_request && cond
        first_active_request.should_update_eligibility = true
      end
    end
    
    def check_requests_tax_revenues_requests_tax_revenues
      first_active_request.tax_revenues_will_change = true if first_active_request

      hrs = Hermes::HousingRequest.where(applicant_ids: id.to_s, is_active: true)
      hrs.each do |hr|
        hr.update_tax_revenues
        hr.save if hr.reference_tax_revenues_changed?
      end
    end

    def coapplicant_first_active_request
      @coapplicant_first_active_request ||= Hermes::HousingRequest.order_by(created_at: :desc).find_by(is_active: true, applicant_ids: self.id)
    end 
    
    def update_first_active_request
      first_active_request&.reload
      prepare_first_active_request_data_changes
      check_monthly_revenues_dependency
      check_category_dependency
      check_eligibility_dependency
      check_requests_tax_revenues_requests_tax_revenues
      if main_employer_siret_changed?
        first_active_request.update_entity_relationship if first_active_request
        if coapplicant_first_active_request
          coapplicant_first_active_request.update_entity_relationship
          coapplicant_first_active_request.save!
        end 
      end 
      if first_active_request&.changed? || should_update_eligibility
        first_active_request&.save!
      end
    end
    
    # def should_touch_active_request
    #   return unless first_active_request
    
    #   first_active_request.monthly_revenues_will_change ||
    #     first_active_request.category_may_change ||
    #     first_active_request.should_update_eligibility ||
    #     first_active_request.name_has_changed ||
    #     first_active_request.tax_revenues_will_change
    # end
    
    def main_user
      User.find_by(id: main_user_id)
    end
    
    def identity_values_changed?
      first_name_changed? || last_name_changed? || email_changed?
    end
    
    def needs_first_identity_set?
      new_record? && !identity_values_changed?
    end
    
    def full_name
      [first_name, last_name].join(' ')
    end
    
    def update_status_dependencies
      self.status_updated_at = DateTime.current
      self.old_status = status_was
    end

    def before_create_actions
      self.writers = [user_id.to_s] if user_id
    end 

    def before_actions
      self.main_user_id = user_id if main_user_id.blank?
      update_birth_dependencies if birth_date_changed?
      self.should_update_eligibility = true
    end

    def update_housing_request_folders_if_needed
      is_needed = (nationality_changed? ||
        work_contract_type_changed? ||
        monthly_incomes_changed? ||
        monthly_charges_changed? ||
        family_situation_changed? ||
        actual_housing_details_changed? ||
        amount_year_minus_one_changed? ||
        amount_year_minus_two_changed?
      
      )
      return unless is_needed
      
      hrs = HousingRequest.any_of({main_applicant_id: id}, {main_applicant_id: id.to_s}, {applicant_ids: id}, applicant_ids: id.to_s)
      
      hrs.each do |concerned_hr|
        concerned_hr.housing_request_folder.compute_required_documents
        concerned_hr.housing_request_folder.save
      end
    end
    
    def prepare_first_active_request_data_changes
      if first_name_changed? ||
        last_name_changed? ||
        email_changed?
        first_active_request&.name_has_changed = true
      end
      
      first_active_request&.birth_date_has_changed = true if birth_date_changed?
    end
    
    def update_required_documents_user_id
      required_documents = Hermes::RequiredDocument.where(applicant_id: id)
      required_documents.each do |rd|
        rd.writers.push(user_id)
        rd.save
      end
    end
    

    def xml_must_be_sended(xml)
      if (last_xml_flux_36 && xml)
        parsed_xml = Hash.from_xml(xml.gsub("\n", ""))
        parsed_last_xml_flux_36 = Hash.from_xml(xml.xml.gsub("\n", ""))
        return parsed_xml.to_s != parsed_last_xml_flux_36.to_s
      else
        return true
      end
    end

    def update_applicant_used_for_siret_in_request
      hr = Hermes::HousingRequest.find_by(applicant_used_for_siret: self.id)
      if (hr)
        set_applicant_used_for_siret_in_request(hr)
      end
    end

    def set_applicant_used_for_siret_in_request(hr)
      values = {applicant_used_for_siret_firstname: first_name,
      applicant_used_for_siret_lastname: last_name,
      applicant_used_for_siret_fullname: full_name,
      
      applicant_used_for_siret_birth_date: birth_date,
      updated_at: DateTime.current}
      values[:applicant_used_for_siret_email] = email if email
      hr.set(values)
    end

    def after_actions
      check_status_changed
      update_first_active_request
      set(should_update_eligibility: nil, updated_at: DateTime.current)
      update_housing_request_folders_if_needed
      update_required_documents_user_id if user_id_changed?
      update_hr_monthly_revenues_coapplicant if monthly_incomes_changed? || monthly_charges_changed?
      update_applicant_used_for_siret_in_request if identity_values_changed?
      true
    end

    def is_same_of?(reverse_mapped_applicant, strict = true)
      # In strict mode, all field need to be same
      if strict
        return reverse_mapped_applicant[:first_name]&.downcase == self.first_name&.downcase &&
        reverse_mapped_applicant[:last_name]&.downcase == self.last_name&.downcase &&
        DateTime.parse(reverse_mapped_applicant[:birth_date])&.to_date == self.birth_date.to_date
      end

      # In not strict mode, need to have two on three fields matching
      [
        reverse_mapped_applicant[:first_name]&.downcase == self.first_name&.downcase,
        reverse_mapped_applicant[:last_name]&.downcase == self.last_name&.downcase,
        DateTime.parse(reverse_mapped_applicant[:birth_date])&.to_date == self.birth_date.to_date
      ].count(true) >= 2
    end

    def self.find_or_create_by_nud(user, unique_number, birth_date = nil)
      # birth_date format: "2020-04-08"
      nud = Hermes::Nud.find_by(unique_number: unique_number)
      return nud.applicant if nud
      
      hri = Hermes::HousingRequestInitializer.find_or_create_by(unique_number: unique_number) do |new_hri|
        # Enter here only if a new record
        applicant = self.new
        new_hri.save
        new_hri.get_xml_from_sne(user, unique_number)
        unless new_hri.xml&.xml
          applicant.errors.add(:nud, :invalid_unique_number)
          return applicant
        end
        
        parsed_xml = Hash.from_xml(new_hri.xml.xml.gsub("\n", ""))
        main_applicant_data = parsed_xml["interfaceNuu"]["demande"]["demandeLogement"]["personnePhysique"]["demandeur"]
        if main_applicant_data['dateNaissance'] != birth_date
          applicant.errors.add(:birth_date, :invalid_birth_date)
          return applicant
        end
        
        applicant = self.find_or_create_by(email: main_applicant_data["mel"])
        unless applicant.user
          user = ::User.create(
            email: main_applicant_data["mel"],
            first_name: main_applicant_data['prenom'],
            last_name: main_applicant_data['nom'],
            auth_roles: ['demandeur'],
            password: SecureRandom.hex(8)
          )
          applicant.update(user: user)
        end
        new_hri.hydrate_housing_request_from_sne(applicant.user.id)
      end
      
      hri.housing_request.main_applicant
    end

    # This method must be called by Batch model (on core => /admin/batchs) to be scheduled
    def self.daily_offers_notifier
      # Get active requests to get active applicants
      Hermes::HousingRequest.where(is_active: true).batch_size(5).each_with_index do |hr, i|
        applicant = hr.main_applicant
        next unless applicant

        position = i+1
        puts "#{position} - Active applicant #{applicant.id}"
  
        # build today and yesterday time at 6:00am
        today_at_six      = (Date.today.beginning_of_day + 6.hours)
        yesterday_at_six  = today_at_six - 1.day
  
        # get eligibility query and merge filter to get only eligible offers created between yesterday and today
        query = applicant.eligibility_query.json_to_h.merge({ 
          published: true,
          created_at: { 
            :$gte => yesterday_at_six,
            :$lte => today_at_six 
          }
        })

        puts "Query => #{query}"

        # Get offer ids and notify applicant with them
        offer_ids = Hermes::HousingOffer.where(query).only(:id).pluck(:id)
        puts "#{offer_ids.count} new eligible offer(s)"
        if offer_ids.count >= 1
          applicant.notify(
            key: NOTIFICATION_KEYS[:applicant][:new_eligible_offers],
            channels: ['email'],
            datas: { offer_ids: offer_ids }
          )
        end
      end
    end
  end
end
